import json
import logging
from pathlib import Path
from flask import Blueprint, Response, jsonify, request
from patrolify.globals import g
from patrolify.consts import RESULT_FILE_NAME
from patrolify.queue_jobs import trigger_target
from patrolify.reports import (
    get_check_report,
    get_latest_check_ids,
    get_result_by_job_id,
)
from rq import Worker


logger = logging.getLogger(__name__)
admin_api_blueprint = Blueprint("admin_api", __name__, url_prefix="/api/v1")


@admin_api_blueprint.route("/triggers")
def triggers_list():
    checker_queue = g.checker_queue
    scheduled_jobs = checker_queue.scheduled_job_registry

    list_of_job_instances = g.scheduler.get_jobs()

    scheduled_triggers = list(
        {"name": j.args[0], "description": j.args[1]} for j in list_of_job_instances
    )

    checkers = []
    for trigger in scheduled_triggers:
        report_path = get_latest_report_dir(trigger["name"])
        if not report_path:
            checkers.append({
                "name": trigger["name"],
                "description": trigger["description"],
                "latest_report_dir": None,
                "latest_report_timestamp": None,
                "report": None,
            })
            continue
        checkers.append({
            "name": trigger["name"],
            "description": trigger["description"],
            "latest_report_dir": str(report_path),
            "latest_report_timestamp": str(report_path.name),
            "report": get_report_by_path(report_path),
        })

    return jsonify({
        "scheduled_jobs_count": scheduled_jobs.count,
        "checkers": checkers,
    })


@admin_api_blueprint.route("/checker/<name>")
def checker_detail(name):
    latest_check_ids = [str(x).strip() for x in get_latest_check_ids(name)]

    result = {}
    for check_id in latest_check_ids:
        result[check_id] = get_check_report(name, check_id)

    return jsonify(result)


@admin_api_blueprint.route("/checker/<name>/latest-result")
def checker_latest_result(name):
    latest_check_ids = get_latest_check_ids(name)
    latest_one = latest_check_ids[-1].strip()
    result = get_check_report(name, latest_one)
    return jsonify(result)


@admin_api_blueprint.route("/checker/<name>/<int:check_id>")
def result_by_check_id(name, check_id):
    return jsonify(get_check_report(name, str(check_id)))


@admin_api_blueprint.route("/checker/<name>/<int:check_id>/<job_id>")
def result_by_job_id(name, check_id, job_id):
    return jsonify(get_result_by_job_id(name, str(check_id), job_id))


@admin_api_blueprint.route("/checker/<name>/enqueue", methods=["POST"])
def enqueue_checker(name):
    result = g.checker_queue.enqueue(trigger_target, name, True)
    logger.info("enqueue a job now, result: %s", result)
    return jsonify({})


@admin_api_blueprint.route("/file")
def get_file():
    path = request.args.get("path")
    logger.info("Get file path=%s", path)
    if not path:
        return Response("path in query params is requied", status=400)

    if path.startswith("checkers"):
        target = Path(g.checker_path) / path.removeprefix("checkers").removeprefix("/")
    elif path.startswith("results"):
        target = Path(g.result_path) / path.removeprefix("results").removeprefix("/")
    else:
        return Response(
            f"path {path} is not allowed, only checkers path checkers and"
            " result path results is allowed",
            status=403,
        )

    if not target.exists():
        return Response(f"{target} do not exist. {g.checker_path}", status=404)

    if target.is_dir():
        return jsonify({
            "type": "directory",
            "files": [{"name": x.name, "is_dir": x.is_dir()} for x in target.iterdir()],
        })
    if target.is_file():
        with open(target) as f:
            return jsonify({"type": "file", "content": f.read()})

    return Response("unsupported type", status=400)


@admin_api_blueprint.route("/monitor-info")
def monitor_info():
    workers = Worker.all(connection=g.redis)

    workers_data = [
        {
            "hostname": worker.hostname,
            "pid": worker.pid,
            "queues": worker.queue_names(),
            "state": worker.state,
            "last_heartbeat": worker.last_heartbeat,
            "birth_date": worker.birth_date,
            "successful_job_count": worker.successful_job_count,
            "failed_job_count": worker.failed_job_count,
            "total_working_time": worker.total_working_time,
        }
        for worker in workers
    ]

    info = g.redis.info()

    return jsonify({
        "total_worker_count": len(workers),
        "workers": workers_data,
        "checker_queue": {
            "count": g.checker_queue.count,
            "started_job": g.checker_queue.started_job_registry.count,
            "finished_job": g.checker_queue.finished_job_registry.count,
            "failed_job": g.checker_queue.failed_job_registry.count,
        },
        "reporter_queue": {
            "count": g.reporter_queue.count,
            "started_job": g.reporter_queue.started_job_registry.count,
            "finished_job": g.reporter_queue.finished_job_registry.count,
            "failed_job": g.reporter_queue.failed_job_registry.count,
        },
        "redis": {"used_memory_human": info["used_memory_human"]},
    })


def get_latest_report_dir(trigger_name):
    report_dir = g.report_base_dir(trigger_name)

    if not report_dir.exists():
        return None
    report_dirs = sorted(report_dir.iterdir())
    logger.info("year list: %s", report_dirs)

    report_dir = report_dirs[-1]
    report_dirs = sorted(report_dir.iterdir())
    logger.info("month list: %s", report_dirs)

    report_dir = report_dirs[-1]
    report_dirs = sorted(report_dir.iterdir())
    logger.info("day list: %s", report_dirs)

    report_dir = report_dirs[-1]
    report_dirs = sorted(report_dir.iterdir())
    logger.info("hour list: %s", report_dirs)

    report_dir = report_dirs[-1]
    report_dirs = sorted(report_dir.iterdir())
    logger.info("ts list: %s", report_dirs)

    report_dir = report_dirs[-1]
    logger.info("latest report dir is: %s", report_dir)

    return report_dir


def get_report_by_path(path):
    with open(path / RESULT_FILE_NAME) as f:
        return json.load(f)



import dataclasses
import glob
import json
import json
import logging
import shutil
from typing import Dict, List

from redis.lock import Lock

from .consts import RESULT_FILE_NAME
from .globals import g

logger = logging.getLogger(__name__)
MAX_LATEST_CHECK_ID_RECORDING = 100


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)


@dataclasses.dataclass()
class MainResult:
    passed: List[str] = dataclasses.field(default_factory=list)
    not_passed: List[str] = dataclasses.field(default_factory=list)
    failed: List[str] = dataclasses.field(default_factory=list)
    all_passed: bool = True
    job_info: Dict[str, Dict] = dataclasses.field(default_factory=dict)
    start_job_id: str | None = None
    manually_triggered: bool = False

def generate_report(check_name, check_id):
    logger.info("Start to generate reports for %s check_id=%s", check_name, check_id)
    check_directory = g.result_dir(check_name, check_id)
    report_directory = g.report_dir(check_name, check_id)

    main_result = MainResult()

    for file in glob.glob(str(check_directory / "*-*-*-*-*.json")):
        logger.debug("Parse result %s...", file)
        with open(file, "r") as f:
            result = json.load(f)

        # save job information
        job_id = result["job_id"]
        target_name = result["target"]
        job_info = main_result.job_info.setdefault(job_id, {})
        job_info["target"] = target_name
        job_info["run_success"] = result['run_success']
        job_info["check_pass"] = result['check_pass']
        job_info.setdefault("child_job_ids", [])

        # save job's relation
        parent_id = result["parent_target_id"]
        if parent_id is None:
            main_result.start_job_id = job_id
            main_result.manually_triggered = result.get("manually_triggered", False)
        else:
            parent = main_result.job_info.setdefault(parent_id, {})
            parent.setdefault("child_job_ids", []).append(job_id)

        # save result
        if not result["run_success"]:
            main_result.failed.append(result["job_id"])
            main_result.all_passed = False
            continue

        if result["check_pass"]:
            main_result.passed.append(result["job_id"])
        else:
            main_result.all_passed = False
            main_result.not_passed.append(result["job_id"])

    check_directory.mkdir(parents=True, exist_ok=True)

    main_result_location = check_directory / RESULT_FILE_NAME
    with open(main_result_location, "w") as f:
        json.dump(main_result, f, cls=EnhancedJSONEncoder)

    shutil.move(check_directory, report_directory)

    logger.info("result has been writen to %s", report_directory)
    update_latest_check(check_name, check_id)


def update_latest_check(check_name, check_id):
    """
    write the check_id into the latest check file
    """
    lock = Lock(g.redis, "patrolify:lock:update_latest_check_file")
    with lock:
        old_check_ids = get_latest_check_ids(check_name)

        existing_records = len(old_check_ids)
        if existing_records >= MAX_LATEST_CHECK_ID_RECORDING:
            start = existing_records - MAX_LATEST_CHECK_ID_RECORDING + 1
            old_check_ids = old_check_ids[start:]

        old_check_ids.append(f"{check_id}\n")
        logger.info("old check ids: %s", old_check_ids)
        with open(g.latest_check_ids_file(check_name), "w") as f:
            f.writelines(old_check_ids)


def get_latest_check_ids(check_name) -> List[str]:
    if not g.latest_check_ids_file(check_name).exists():
        return []
    with open(g.latest_check_ids_file(check_name), "r") as f:
        old_check_ids = f.readlines()
    return old_check_ids


def get_check_report(check_name, check_id):
    with open(g.report_dir(check_name, check_id) / RESULT_FILE_NAME) as f:
        report_result = json.load(f)

    return report_result


def get_result_by_job_id(check_name, check_id, job_id):
    with open(g.report_dir(check_name, check_id) / f"{job_id}.json") as f:
        report_result = json.load(f)

    return report_result



def store_check_result(result, target, check_name, check_id, job_id, parent_id, manually_triggered):
    data = {
        "job_id": job_id,
        "target": target,
        "parent_target_id": parent_id,
        "run_success": None,
        "check_pass": None,
        "reason": None,
    }

    # only store this when it's true
    if manually_triggered:
        data['manually_triggered'] = manually_triggered

    if result is None:
        pass
    elif isinstance(result, tuple):
        boolresult, reason = result

        data.update(
            {
                "run_success": True,
                "check_pass": boolresult,
                "reason": reason,
            }
        )

    result_dir = g.result_dir(check_name, check_id)
    result_dir.mkdir(parents=True, exist_ok=True)
    with open(str(result_dir / job_id) + ".json", "w") as f:
        json.dump(data, f)
        logger.info("Job result has been saved to %s", f)



from datetime import datetime
import logging

from patrolify.exceptions import UsageError

from .globals import Role, g
from .queue_jobs import trigger_target


logger = logging.getLogger(__name__)


def check(target_cls):
    def wrapper(func):
        logger.info("Register checker for target: %s", target_cls)
        # register the target checker
        g.target_checkers[target_cls] = func
        return func

    return wrapper


def trigger(interval_seconds=None, cron_string=None, description=""):
    if not interval_seconds and not cron_string:
        raise UsageError("Must set interval_seconds or cron_string")

    scheduler = g.scheduler

    def wrapper(func):
        funcname = f"{func.__module__}.{func.__name__}"
        if g.role != Role.PLANNER:
            g.triggers[funcname] = func
            logger.info("Regisgered a trigger: %s", funcname)
            return func
        logger.info(
            "Schedule jobs func %s for interval_seconds=%s, cron_string=%s",
            funcname,
            interval_seconds,
            cron_string,
        )

        if interval_seconds:
            scheduler.schedule(
                scheduled_time=datetime.utcnow(),
                func=trigger_target,
                interval=interval_seconds,
                args=[funcname, False, description],
            )
        else:
            scheduler.cron(
                cron_string,
                func=trigger_target,
                args=[funcname, False, description],
            )
        # TODO schedule jobs
        return func

    return wrapper



from datetime import datetime
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
import threading
from typing import Callable, Dict
from .consts import LATEST_CHECK_FILE_NAME


class Role(Enum):
    SCHEDULER = "sheduler"
    PLANNER = "planner"
    WORKER = "worker"
    REPORTER = "reporter"


@dataclass
class Global:
    target_checkers: Dict = field(default_factory=dict)
    triggers: Dict[str, Callable] = field(default_factory=dict)
    checker_queue = None
    reporter_queue = None
    scheduler = None
    role: Role = Role.WORKER
    redis = None
    result_path = None
    checker_path = None

    def result_dir(self, check_name, check_id):
        """save jobs result"""
        return Path(self.result_path) / check_name / "results" / check_id

    def report_dir(self, check_name, check_id):
        """save a check's result, when all jobs are finished, the result will
        be moved from ``result_dir`` to ``report_dir``
        """
        # year, month, day, hour
        time_path = datetime.utcfromtimestamp(int(check_id)).strftime("%Y/%m/%d/%H")
        return self.report_base_dir(check_name) / time_path / check_id

    def report_base_dir(self, check_name):
        return Path(self.result_path) / check_name / "reports"

    def latest_check_ids_file(self, check_name):
        return Path(self.result_path) / check_name / LATEST_CHECK_FILE_NAME


g = Global()

threadlocal = threading.local()



from datetime import datetime
import logging
import time
import uuid

from .globals import threadlocal


logger = logging.getLogger(__name__)


class Target:
    def __init__(self) -> None:
        self.check_id = threadlocal.check_id
        self.check_name = threadlocal.check_name

        self.job_id = str(uuid.uuid4())
        if hasattr(threadlocal, "current_target"):
            self.parent_target = threadlocal.current_target.job_id
        else:
            self.parent_target = None

    @property
    def scheduled_count_key(self):
        return f"patrolify:{self.check_name}:{self.check_id}:scheduled"

    @property
    def finished_count_key(self):
        return f"patrolify:{self.check_name}:{self.check_id}:finished"

    @property
    def started_time_key(self):
        return f"patrolify:{self.check_name}:{self.check_id}:started_timestamp"


class TimeTriggerTarget(Target):
    def __init__(self) -> None:
        super().__init__()
        self.trigger_timestamp = time.time()
        self.manually_triggered = False

    def __str__(self):
        datetime_obj = datetime.fromtimestamp(self.trigger_timestamp)
        formatted_string = datetime_obj.strftime("%Y-%m-%d %H:%M:%S")
        manual = ""
        if self.manually_triggered:
            manual = " (manually triggered)"
        return f"Trigger at {formatted_string} {manual}"



import logging
import time
import types
from patrolify.consts import GLOBAL_TTL_SECONDS

from patrolify.target import TimeTriggerTarget

from .globals import g, threadlocal
from .reports import generate_report, store_check_result

logger = logging.getLogger(__name__)


def check_target(target):
    logger.info("check on target=%s", target)
    restore_thread_local(target)

    checker_func = get_checker(target)
    result = checker_func(target)
    return process_result(result, target)


def trigger_target(funcname, manually_triggered=False, description=""):
    threadlocal.check_id = str(int(time.time()))
    threadlocal.check_name = funcname

    logger.info("Start running the trigger function: %s", funcname)
    target = TimeTriggerTarget()
    if manually_triggered:
        target.manually_triggered = True

    redis = g.redis
    exat = int(time.time() + GLOBAL_TTL_SECONDS)
    redis.set(target.scheduled_count_key, 1, exat=exat)
    redis.set(target.finished_count_key, 0, exat=exat)
    redis.set(target.started_time_key, time.time(), exat=exat)

    threadlocal.current_target = target
    trigger = g.triggers[funcname]
    result = trigger(target)

    process_result(result, target)


def process_result(result, target):
    manually_triggered = False
    if hasattr(target, "manually_triggered"):
        manually_triggered = target.manually_triggered
    if isinstance(result, tuple):
        g.reporter_queue.enqueue(
            store_check_result,
            result,
            str(target),
            target.check_name,
            target.check_id,
            target.job_id,
            target.parent_target,
            manually_triggered,
        )

    if isinstance(result, types.GeneratorType):
        try:
            while True:
                t = next(result)
                queue_target(t)
        except StopIteration as e:
            check_result = e.value
            g.reporter_queue.enqueue(
                store_check_result,
                check_result,
                str(target),
                target.check_name,
                target.check_id,
                target.job_id,
                target.parent_target,
                manually_triggered,
            )

    incr_task_count_and_check_finsihed(target)


def incr_task_count_and_check_finsihed(target):
    redis = g.redis

    finished_count = redis.incr(target.finished_count_key)
    scheduled_count = int(redis.get(target.scheduled_count_key))

    if finished_count == scheduled_count:
        logger.info(
            "all tasks has been finsiehd, %s=%d, %s=%d",
            target.scheduled_count_key,
            scheduled_count,
            target.finished_count_key,
            finished_count,
        )
        g.reporter_queue.enqueue(generate_report, target.check_name, target.check_id)


def get_checker(target):
    target_type = target.__class__
    checker_func = g.target_checkers[target_type]
    return checker_func


def queue_target(target):
    redis = g.redis
    redis.incr(target.scheduled_count_key)
    g.checker_queue.enqueue(check_target, target)


def restore_thread_local(target):
    threadlocal.check_id = target.check_id
    threadlocal.check_name = target.check_name
    threadlocal.current_target = target



import sys
import importlib
import waitress
import importlib.util
import logging
import os
import pathlib

import click
from redis import Redis
from rq import Queue, Worker

from rq_scheduler import Scheduler

from .globals import Role, g
from .reports import generate_report
from .admin import create_app
from patrolify import __version__


logger = logging.getLogger(__name__)
LOG_LEVEL = {
    0: logging.CRITICAL,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}


def setup_log(enabled, level, loglocation):
    if enabled:
        logging.basicConfig(
            filename=os.path.expanduser(loglocation),
            filemode="a",
            format="%(asctime)s %(levelname)5s (%(module)sL%(lineno)d) %(message)s",
            level=level,
        )
    else:
        logging.disable(logging.CRITICAL)
    logger.info("------ patrolify ------")


def load_checkers(path):
    sys.path.append(str(pathlib.Path(path).parent))
    for root, _, files in os.walk(path):
        for file in files:
            abs_path = os.path.join(root, file)
            if not file.endswith(".py"):
                logger.debug("%s is not ending with .py, ignore...", abs_path)
                continue
            logger.info("import python file: %s", abs_path)

            # remove the ./ prefix
            abs_path = abs_path.lstrip("./")

            module_name = abs_path[:-3].replace("/", ".")
            logger.info("loading module: %s", module_name)
            importlib.import_module(module_name)
    logger.info("Loading checkers done: %s", g.target_checkers)


def print_version(ctx, _, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(__version__)
    ctx.exit()


@click.group()
@click.option(
    "-v",
    "--verbose",
    count=True,
    default=2,
    help="Add log verbose level, using -v, -vv, -vvv for printing more logs.",
)
@click.option(
    "-l",
    "--log-to",
    type=click.Path(),
    default=None,
    help="Printing logs to a file, for debugging, default is no logs.",
)
@click.option("-r", "--redis-url", help="Redis url as the job queue")
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
)
def main(verbose, log_to, redis_url):
    log_level = LOG_LEVEL[verbose]
    setup_log(log_to is not None, log_level, log_to)
    g.redis = Redis.from_url(redis_url)
    g.checker_queue = Queue("checker", connection=g.redis)
    g.reporter_queue = Queue("reporter", connection=g.redis)


@main.command()
@click.option(
    "-p",
    "--python-checker",
    type=click.Path(),
    default=None,
    help="Python checker script location",
)
@click.option(
    "-r",
    "--result-path",
    type=click.Path(),
    default=None,
    help="Where to save the results",
)
@click.option(
    "--queue",
    "-q",
    multiple=True,
    help=(
        "work for which queue? queue type: 1) checker 2) patrolify, for generating"
        " reports. If you use the file system as the report store, you should only run"
        " one worker for the patrolify queue."
    ),
)
def worker(python_checker, result_path, queue):
    queue_map = {
        "checker": g.checker_queue,
        "reporter": g.reporter_queue,
    }

    _queue = []
    for item in queue:
        if item not in queue_map:
            raise click.UsageError(
                f"Queue {item} not in {','.join(queue_map.keys())}", ctx=None
            )
        _queue.append(queue_map[item])

    g.role = Role.WORKER
    g.result_path = pathlib.Path(result_path)
    load_checkers(python_checker)
    start_worker(_queue)


@main.command()
@click.option(
    "-p",
    "--python-checker",
    type=click.Path(),
    default=None,
    help="Python checker script location",
)
def planner(python_checker):
    g.role = Role.PLANNER
    g.scheduler = Scheduler(
        queue=g.checker_queue, connection=g.checker_queue.connection
    )

    # cancel the existing jobs
    list_of_job_instances = g.scheduler.get_jobs()
    for j in list_of_job_instances:
        g.scheduler.cancel(j)
        logger.info("job %s has been canceled", j)

    load_checkers(python_checker)


@main.command(help="Manually generate reports")
@click.option("--check-name", help="The check job's name")
@click.option("--check-id", help="The ID of one check, it's normally an int timestamp")
@click.option(
    "-r",
    "--result-path",
    type=click.Path(),
    default=None,
    help="Where to save the results",
)
def generate_reports(check_name, check_id, result_path):
    g.role = Role.REPORTER
    g.result_path = result_path
    logger.info(
        "Manually trigger generate report for job=%s, check_id=%s", check_name, check_id
    )
    generate_report(check_name, check_id)


@main.command(help="Admin web HTTP server")
@click.option("--host", default="127.0.0.1", help="The interface to bind to.")
@click.option("--port", default=8080, help="The port to bind to.")
@click.option("--connection-limit", "-c", default=1000, help="Server connection limit")
@click.option("--threads", "-t", default=64, help="Server threads")
@click.option(
    "-p",
    "--python-checker",
    type=click.Path(),
    default=None,
    help="Python checker script location",
    required=True,
)
@click.option(
    "-r",
    "--result-path",
    type=click.Path(),
    default=None,
    required=True,
    help="Where to save the results",
)
@click.option("--control-api/--no-control-api", default=True)
def admin(
    host, port, connection_limit, threads, python_checker, result_path, control_api
):
    load_checkers(python_checker)
    g.result_path = result_path
    g.checker_path = python_checker
    g.scheduler = Scheduler(
        queue=g.checker_queue, connection=g.checker_queue.connection
    )

    app = create_app(control_api=control_api)
    waitress.serve(
        app,
        host=host,
        port=port,
        connection_limit=connection_limit,
        threads=threads,
    )


def start_worker(queue):
    w = Worker(queue, connection=g.redis)
    w.work()



import logging


from flask import Flask
from werkzeug.middleware.dispatcher import DispatcherMiddleware
from prometheus_client import make_wsgi_app
from .admin_api import admin_api_blueprint


logger = logging.getLogger(__name__)


def create_app(control_api):
    app = Flask(__name__, static_folder="frontend_dist", static_url_path="/static")

    app.config.update(ENABLE_CONTROL_API=control_api)

    # Add prometheus wsgi middleware to route /metrics requests
    prometheus_wsgi_app = make_wsgi_app()
    app.wsgi_app = DispatcherMiddleware(
        app.wsgi_app,
        {
            "/metrics": prometheus_wsgi_app,
        },
    )
    app.register_blueprint(admin_api_blueprint)

    @app.route("/", defaults={"path": ""})
    @app.route("/<path:path>")
    def catch_all(path):
        return app.send_static_file("index.html")

    return app
